from typing import Iterator, Optional, Any

from fpylib.irange.num_irange import (
    _generate_finite_num_range,
    _generate_infinite_num_range,
)


def is_char(*objs: Any, allow_none=False) -> bool:
    """
    Checks if the object is a character.

    :obj: The object to be checked.
    :return: True if the object is a character, False otherwise.
    """
    return all(
        (isinstance(obj, str) and len(obj) == 1)
        or (allow_none and (obj is None or obj is Ellipsis))
        for obj in objs
    )


def char_encode(num: int, is_lower=True) -> str:
    """
    Converts an integer to a character string \
    with elements from a-z or A-Z.

    :num: The integer base 10 to be converted.
    :return: The character base 26.
    """
    buffer = 97 if is_lower else 65
    result = "" if num != 0 else chr(buffer + num % 26)
    while num > 0:
        result = chr(buffer + num % 26) + result
        num //= 26
    return result


def char_decode(string: str) -> int:
    """
    Converts an string base 26 to a number of base 10.

    :string: The string to be converted.
    :return: The number base 10.
    """
    if not all(is_char(char) for char in string):
        raise TypeError("The string must be a character [a-zA-Z]")

    if 97 <= ord(string[0]) <= 122:
        buffer = 97
    elif 65 <= ord(string[0]) <= 90:
        buffer = 65
    else:
        raise ValueError(f"{string} must be characters [a-zA-Z]")

    return sum((ord(char) - buffer) * (26 ** i) for i, char in enumerate(string[::-1]))


def is_lower_char(obj: Any) -> bool:
    """
    Checks if the object is a letter.

    :obj: The object to be checked.
    :return: True if is a lower character, \
    False if is upper character and raises TypeError otherwise.
    """
    if not is_char(obj):
        raise TypeError("The object must be a character [a-zA-Z]")
    elif ord(obj) >= 97 and ord(obj) <= 122:
        return True
    return False


def _generate_infinite_char_range(first: str, second: Optional[str]) -> Iterator[str]:
    """
    Generates an infinite range of characters.

    :first: The first character of the range.
    :second: The second character of the range.
    :return: A generator of the range.
    """
    is_lower = is_lower_char(first)
    _first, _second = (
        char_decode(string) if string else None for string in (first, second)
    )
    yield from (
        char_encode(i, is_lower) for i in _generate_infinite_num_range(_first, _second)
    )


def _generate_finite_char_range(
    first: str, second: Optional[str], final: Optional[str], final_include: bool
) -> Iterator[str]:
    """
    Generates a finite range of characters.

    :first: The first character of the range.
    :second: The second character of the range.
    :final: The final character of the range.
    :final_include: True if the final character is included in the range.
    :return: A generator of the range.
    """
    is_lower = is_lower_char(first)
    _first, _second, _final = (
        char_decode(string) if string else None for string in (first, second, final)
    )

    yield from (
        char_encode(i, is_lower)
        for i in _generate_finite_num_range(_first, _second, _final, final_include)
    )


def char_irange(
    first: str,
    second: Optional[str] = None,
    final: Optional[str] = None,
    final_include: bool = False,
) -> Iterator[str]:
    """
    Intelligen range function that can be used \
    with infinite or finite ranges of characters.

    :first: The first character of the range.
    :second: The second character of the range.
    :final: The final character of the range.
    :return: A generator of the range.
    """

    if isinstance(final, str):
        yield from _generate_finite_char_range(first, second, final, final_include)
    elif final in [None, Ellipsis]:
        yield from _generate_infinite_char_range(first, second)
    else:
        raise TypeError("Final argument must be an character")



from fpylib.composer import compose, parallelize
from fpylib.lazyness import lazy_eval, lazy_class
from fpylib.irange import irange, char_encode, char_decode, num_irange, char_irange
from fpylib.functors import *



from typing import Any, Callable, Generic, Optional
from fpylib.functors.functor import Functor, T, S, fmap
from fpylib.functors.monad import Monad
from fpylib.lazyness import lazy_eval


class Applicative(Functor, Generic[T]):
    def apply(self, func: "Functor[Callable[[T], S]]") -> "Functor[S]":
        return self.bind(func.get())


@lazy_eval
def apply(func: Functor[Callable[[T], S]], ft: Functor[T]) -> Functor[S]:
    """
    This the function applicative to a functor that wraps a function and a functor.

    :param func: Functor[Callable[[T], S]]
    :param ft: Functor[T]
    :return: Functor[S]
    """
    return ft.apply(func)


def lift_a2(func: Callable[[T, T], S], f1: Functor[T], f2: Functor[T]) -> Functor[S]:
    """
    This is the version Pythonic to liftA2 from Haskell.

    :param func: Callable[[T, T], S]
    :param f1: Functor[T]
    :param f2: Functor[T]
    :return: Functor[S]
    """
    return apply(fmap(lazy_eval(func), f1), f2)



from fpylib.functors.functor import fmap
from fpylib.functors.applicative import apply, lift_a2
from fpylib.functors.monad import unit
from fpylib.functors.maybe import Maybe, Just, Nothing
from fpylib.functors.flist import *



from typing import Callable, Any, Tuple
from functools import wraps


def compose(*funcs: Callable[..., Any]) -> Callable[..., Any]:
    """
    Composes two or more functions.

    :funcs: The functions to be composed.
    :return: The composition of the two functions.
    """
    if not funcs:
        raise ValueError("No functions to compose")
    elif len(funcs) == 1:
        return funcs[0]

    @wraps(funcs)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        result = funcs[0](*args, **kwargs)
        for func in funcs[1:]:
            result = func(result)
        return result

    return wrapper


def parallelize(*funcs: Callable[..., Any]) -> Callable[..., Tuple]:
    """
    Decorator to make a function parallelizable.

    :func: The function to be parallelizable.
    :return: A function that returns the result of the function.
    """

    @wraps(funcs)
    def wrapper(*args: Any, uniqui_intput: bool = True) -> Tuple:
        if not uniqui_intput:
            return tuple(func(arg) for arg, func in zip(args, funcs))
        return tuple(func(*args) for func in funcs)

    return wrapper



from typing import Iterator, Optional, Union
from numbers import Number

from fpylib.irange.num_irange import num_irange
from fpylib.irange.char_irange import char_irange


def irange(
    first: Union[Number, str],
    second: Optional[Union[Number, str]] = None,
    final: Optional[Union[Number, str]] = None,
    final_include: bool = False,
) -> Iterator[Union[Number, str]]:
    """
    Intelligen range function that can be used with infinite or finite ranges.

    :first: The first number of the range.
    :second: The second number of the range.
    :final: The final number of the range.
    :return: A generator of the range.
    """

    if isinstance(first, Number):
        yield from num_irange(first, second, final, final_include)
    elif isinstance(first, str):
        yield from char_irange(first, second, final, final_include)
    else:
        raise TypeError(f"Invalid type of {first}")



from typing import Callable, Generic, Optional
from fpylib.functors.functor import T, S, Functor
from fpylib.functors.monad import Monad
from fpylib.functors.applicative import Applicative


class Maybe(Applicative, Monad, Generic[T]):
    def unit(self, value: T) -> "Maybe[T]":
        if value is not None:
            return Just(value)
        return Nothing()

    def bind(self, func: Callable[[T], S]) -> "Maybe[S]":
        try:
            value = func(self.get())
            if value is None:
                return Nothing(ValueError("The value is None"))
            return Just(value)
        except Exception as e:
            return Nothing(e)


class Just(Maybe):
    def __str__(self) -> str:
        return f"Just {self.get()}"

    def __repr__(self) -> str:
        return f"Just {type(self.get())}"


class Nothing(Maybe):
    def __init__(self, *failure: Optional[Exception]) -> None:
        """
        This does nothing.
        """
        object.__setattr__(
            self, "_Nothing__failure", filter(lambda fail: fail is not None, failure)
        )

    def fails(self) -> bool:
        return self.__failure

    def __str__(self) -> str:
        return "Nothing"

    def __repr__(self) -> str:
        return f"{self.__str__()} {list(self.__failure)}"



from typing import Iterator, Optional
from numbers import Number


def _number_generator(
    start: Number, final: Optional[Number] = None, step: Number = 1
) -> Iterator[Number]:
    current = start
    while True if final is None else current < final:
        yield current
        current += step


def _generate_finite_num_range(
    first: Number,
    second: Optional[Number],
    final: Optional[Number],
    final_include: bool,
) -> Iterator[Number]:
    if second in [None, Ellipsis]:
        yield from _number_generator(first, final + final_include)
    elif isinstance(second, Number):
        yield from _number_generator(first, final + final_include, second - first)
    else:
        raise TypeError("Second argument must be an integer")


def _generate_infinite_num_range(
    first: Number, second: Optional[Number]
) -> Iterator[Number]:
    if second in [None, Ellipsis]:
        yield from _number_generator(first)
    elif "__sub__" in dir(second):
        yield from _number_generator(first, step=second - first)
    else:
        raise TypeError(f"{type(first)} argument must have __sub__")


def num_irange(
    first: Number,
    second: Optional[Number] = None,
    final: Optional[Number] = None,
    final_include: bool = False,
) -> Iterator[Number]:
    """
    Intelligen range function that can be used \
    with infinite or finite ranges of integers.

    :first: The first character of the range.
    :second: The second character of the range.
    :final: The final character of the range.
    :return: A generator of the range.
    """

    if isinstance(final, Number):
        yield from _generate_finite_num_range(first, second, final, final_include)
    elif final in [None, Ellipsis]:
        yield from _generate_infinite_num_range(first, second)
    else:
        raise TypeError("Final argument must be an integer")



from typing import Any, Callable, Generic, Optional
from fpylib.functors.functor import T, S
from fpylib.lazyness import lazy_eval


class Monad(Generic[T]):
    def __init__(self, value: Optional[T] = None) -> None:
        object.__setattr__(self, "_Monad__value", value)

    def __setattr__(self, __name: str, __value: Any) -> None:
        raise AttributeError("This object is not modifiable")

    def get(self) -> T:
        return self.__value

    def unit(self, value: T) -> "Monad[T]":
        return Monad(value)

    def bind(self, func: Callable[[[T]], S]) -> "Monad[S]":
        return Monad(func(self.get()))

    def __rshift__(self, func: Callable[[[T]], S]) -> "Monad[S]":
        return self.bind(func)


@lazy_eval
def unit(m: "Monad", value: T) -> "Monad[T]":
    """
    The unit function for the Monad.

    :param m: The Monad.
    :param value: The value to be wrapped in the Monad.
    :return: The Monad with the value.
    """
    return m.unit(m, value)



from functools import partial, wraps
from typing import Callable, Union, Any
from inspect import getmembers, isfunction, signature


def lazy_eval(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to make a function lazy.

    :func: The function to be lazy.
    :return: A function that returns the result of the function.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Union[Callable[..., Any], Any]:
        if len(signature(func).parameters) == len(args) + len(kwargs):
            return func(*args, **kwargs)
        elif len(signature(func).parameters) > len(args) + len(kwargs):
            return lazy_eval(partial(func, *args, **kwargs))
        raise ValueError("Too many arguments")

    return wrapper


def lazy_class(cls: object) -> object:
    """
    Decorator to make a class lazy.

    :cls: The class to be lazy.
    :return: A class with all the methods lazies.
    """

    @wraps(cls)
    def wrapper(cls: object) -> Callable[[object], object]:
        for name, method in getmembers(cls, predicate=isfunction):
            setattr(cls, name, lazy_eval(method))
        return cls

    return wrapper(cls)



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    README = fh.read()

setup(
    name="fpylib",
    packages=find_packages(include=["fpylib", "fpylib.irange", "fpylib.functors"]),
    version="0.1.1.2",
    description="This is a library to do functional programming in Python.",
    long_description_content_type="text/markdown",
    long_description=README,
    url="https://github.com/FabianVegaA/fpy_lib",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    author="Fabián Vega",
    license="MIT",
    install_requires=[],
    setup_requires=["pytest-runner"],
    tests_require=["pytest==4.4.1"],
    test_suite="tests",
)



from typing import Any, Callable, Generic, Optional, TypeVar
from fpylib.lazyness import lazy_eval

T = TypeVar("T")
S = TypeVar("S")


class Functor(Generic[T]):
    def __init__(self, value: Optional[T] = None) -> None:
        def frozen_setattr(cls, __name: str, __value: Any) -> None:
            raise AttributeError("This object is not modifiable")

        object.__setattr__(self, "_Functor__value", value)
        object.__setattr__(self, "__setattr__", frozen_setattr)

    def get(self) -> T:
        return self.__value

    def fmap(self, func: Callable[["Functor[T]"], "Functor[S]"]) -> "Functor[S]":
        return Functor(func(self.get()))


@lazy_eval
def fmap(func: Callable[[Functor[T]], Functor[S]], ft: Functor[T]) -> Functor[S]:
    """
    This function is a functor's fmap function.

    :param func: Callable[[Functor[T]], Functor[S]]
    :param ft: Functor[T]
    :return: Functor[S]
    """
    return ft.fmap(func)



from typing import Any, Callable, Generic, Iterable, List, Optional, Tuple
from fpylib.functors.applicative import Applicative
from fpylib.functors.functor import T, S, Functor
from fpylib.functors.monad import Monad
from fpylib.functors.maybe import Just, Maybe, Nothing

from functools import reduce


class FList(Applicative, Monad, Generic[T]):
    def __new__(cls, value: Optional[T] = None) -> "FList[T]":
        if value is None:
            return EmptyFList()
        if not isinstance(value, Iterable):
            value = [value]

        elems = list(filter(lambda x: x is not None, value))

        if not elems:
            return EmptyFList()

        obj = super().__new__(cls)
        object.__setattr__(obj, "_Functor__value", value)
        return obj

    def __init__(self, value: Optional[T] = None) -> None:
        def frozen_setattr(cls, __name: str, __value: Any) -> None:
            raise AttributeError("This object is not modifiable")

        object.__setattr__(self, "__setattr__", frozen_setattr)

    def unit(self, value: Optional[Iterable[T]] = None) -> "FList[T]":
        return self.__new__(self, value=value)

    def bind(self, func: Callable[[T], S]) -> "FList[S]":
        if isinstance(self, EmptyFList):
            return EmptyFList()
        return FList(list(map(func, self)))

    def fmap(self, func: Callable[[T], S]) -> "FList[S]":
        return self.bind(func)

    def apply(self, lfunc: "FList[Callable[[T], S]]") -> "FList[S]":
        return reduce(lambda l1, l2: l1 + l2, [self.bind(func) for func in lfunc])

    def get(self) -> List[T]:
        return super().get()

    def __add__(self, other: "FList[T]") -> "FList[T]":
        return FList(list(self.get()) + list(other.get()))

    def __iter__(self) -> Iterable[T]:
        return self.get().__iter__()

    def __next__(self) -> T:
        return next(self.get())

    def __getitem__(self, __s: slice) -> "FList[T]":
        return self.unit(self.get().__getitem__(__s))

    def __len__(self) -> int:
        return len(self.get())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FList):
            return False
        return self.get() == other.get()

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __contains__(self, item: T) -> bool:
        return item in self.get()

    def __bool__(self) -> bool:
        return not isinstance(self, EmptyFList)

    def __str__(self) -> str:
        return f"FList {self.get()}"

    def __repr__(self) -> str:
        return f"FList {list(set(map(type, self.get())))}"


class EmptyFList(FList):
    def __new__(cls) -> "EmptyFList":
        return Functor.__new__(cls)

    def get(self) -> List[T]:
        return []

    def __str__(self) -> str:
        return "EmptyFList"

    def __repr__(self) -> str:
        return self.__str__()


def concat(*ls: "FList[T]") -> "FList[T]":
    """
    Concatenate two or more FList.

    :param ls: FList to concatenate
    :return: FList
    """
    return FList(sum(ls, EmptyFList()))


def head(l: "FList[T]") -> T:
    """
    Get the first element of a FList.

    :param l: FList
    :return: First element of FList
    """
    if isinstance(l, EmptyFList):
        raise ValueError("head: empty list")
    return l.get()[0]


def last(l: "FList[T]") -> T:
    """
    Get the last element of a FList.

    :param l: FList
    :return: Last element of FList
    """
    if isinstance(l, EmptyFList):
        raise ValueError("last: empty list")
    return l.get()[-1]


def tail(l: "FList[T]") -> "FList[T]":
    """
    Get the all elements of a FList except the first one.

    :param l: FList
    :return: Tail of FList
    """
    if isinstance(l, EmptyFList):
        raise ValueError("tail: empty list")
    return FList(l.get()[1:])


def init(l: "FList[T]") -> "FList[T]":
    """
    Get all elements of a FList except the last one.

    :param l: FList
    :return: Tail of FList
    """
    if isinstance(l, EmptyFList):
        raise ValueError("init: empty list")
    return FList(l.get()[:-1])


def uncons(l: "FList[T]") -> "Maybe[Tuple[T, 'FList[T]']]":
    """
    Get the first element of a FList and the rest of the FList.

    :param l: FList
    :return: First element of FList and the rest of the FList
    """
    if isinstance(l, EmptyFList):
        raise Nothing()
    return Just(l.get()[0], FList(l.get()[1:]))


def singleton(x: T) -> "FList[T]":
    """
    Create a FList with a single element.

    :param x: Element
    :return: FList
    """
    return FList([x])


def null(l: "FList[T]") -> bool:
    """
    Verify if a FList is empty.

    :param l: FList
    :return: True if FList is empty, False otherwise
    """
    if not isinstance(l, FList):
        raise TypeError(f"null: {l} is not a FList")
    return isinstance(l, EmptyFList)


def length(l: "FList[T]") -> int:
    """
    Get the length of a FList.

    :param l: FList
    :return: Length of FList
    """
    return len(l.get())


def reverse(l: "FList[T]") -> "FList[T]":
    """
    Reverse a FList.

    :param l: FList
    :return: Reversed FList
    """
    return FList(reversed(l.get()))
